from typing import Any, Union


PortValue = Union[int, float, bool]
NullablePortValue = Union[int, float, bool, None]
PortValueChoices = list[dict[str, Union[str, int, float]]]

Attribute = Union[int, float, bool, str, list[dict]]
Attributes = dict[str, Attribute]

AttributeDefinition = dict[str, Any]
AttributeDefinitions = dict[str, AttributeDefinition]

GenericJSONDict = dict[str, Any]
GenericJSONList = list[dict[str, Any]]



from .base import EvalContext, EvalResult
from .functions import Function, function


@function('TIME')
class TimeFunction(Function):
    MIN_ARGS = MAX_ARGS = 0
    DEPS = {'second'}

    async def _eval(self, context: EvalContext) -> EvalResult:
        return context.timestamp


@function('TIMEMS')
class TimeMSFunction(Function):
    MIN_ARGS = MAX_ARGS = 0
    DEPS = {'asap'}

    async def _eval(self, context: EvalContext) -> EvalResult:
        return context.now_ms



from .base import EvalContext, EvalResult
from .functions import Function, function


@function('IF')
class IfFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        if eval_args[0]:
            return eval_args[1]
        else:
            return eval_args[2]


@function('EQ')
class EqFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] == eval_args[1])


@function('GT')
class GTFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] > eval_args[1])


@function('GTE')
class GTEFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] >= eval_args[1])


@function('LT')
class LTFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] < eval_args[1])


@function('LTE')
class LTEFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return int(eval_args[0] <= eval_args[1])



from .base import Event, Handler
from .device import DeviceEvent, DeviceUpdate, FullUpdate
from .handlers import cleanup as cleanup_handlers
from .handlers import disable, enable
from .handlers import init as init_handlers
from .handlers import register_handler, trigger
from .port import PortAdd, PortEvent, PortRemove, PortUpdate, ValueChange


async def trigger_full_update() -> None:
    await trigger(FullUpdate())


async def init() -> None:
    await init_handlers()


async def cleanup() -> None:
    await cleanup_handlers()



from .base import EvalContext, EvalResult
from .functions import Function, function


@function('MIN')
class MinFunction(Function):
    MIN_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        m = eval_args[0]
        for e in eval_args[1:]:
            if e < m:
                m = e

        return m


@function('MAX')
class MaxFunction(Function):
    MIN_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        m = eval_args[0]
        for e in eval_args[1:]:
            if e > m:
                m = e

        return m


@function('AVG')
class AvgFunction(Function):
    MIN_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return sum(eval_args) / len(eval_args)



from typing import Optional

from . import TIME_JUMP_THRESHOLD
from .base import EvalContext, EvalResult
from .exceptions import EvalSkipped
from .functions import Function, function


@function('DELAY')
class DelayFunction(Function):
    MIN_ARGS = MAX_ARGS = 2
    DEPS = {'asap'}
    HISTORY_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: list[tuple[int, float]] = []
        self._last_value: Optional[float] = None
        self._current_value: Optional[float] = None

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, delay = await self.eval_args(context)

        if self._current_value is None:
            self._current_value = value

        # Detect value transitions and build history
        if value != self._last_value:
            self._last_value = value

            # Drop elements from queue if history size reached
            while len(self._queue) >= self.HISTORY_SIZE:
                self._queue.pop(0)

            self._queue.append((context.now_ms, value))

        # Process history
        while self._queue and (context.now_ms - self._queue[0][0]) >= delay:
            self._current_value = self._queue.pop(0)[1]

        if self._queue:
            self.pause_asap_eval(self._queue[0][0] + delay)
        else:
            self.pause_asap_eval(context.now_ms + delay)

        return self._current_value


@function('SAMPLE')
class SampleFunction(Function):
    MIN_ARGS = MAX_ARGS = 2
    DEPS = {'asap'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_duration_ms: int = 0
        self._last_time_ms: int = 0

    async def _eval(self, context: EvalContext) -> EvalResult:
        if context.now_ms - self._last_time_ms < self._last_duration_ms:
            self.pause_asap_eval(self._last_time_ms + self._last_duration_ms)
            return self._last_value

        self._last_value, self._last_duration_ms = await self.eval_args(context)
        self._last_time_ms = context.now_ms

        return self._last_value


@function('FREEZE')
class FreezeFunction(Function):
    MIN_ARGS = MAX_ARGS = 2
    DEPS = {'asap'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_duration_ms: int = 0
        self._last_time_ms: int = 0

    async def _eval(self, context: EvalContext) -> EvalResult:
        if self._last_time_ms == 0:  # idle
            value = await self.args[0].eval(context)
            if value != self._last_value:  # value change detected, start timer
                self._last_time_ms = context.now_ms
                self._last_duration_ms = await self.args[1].eval(context)
                self._last_value = value
            else:
                self.pause_asap_eval()
        else:  # timer active
            if context.now_ms - self._last_time_ms > self._last_duration_ms:  # timer expired
                self._last_time_ms = 0
                # Call _eval() again, now that _last_time_ms is 0
                return await self._eval(context)
            else:
                self.pause_asap_eval(self._last_time_ms + self._last_duration_ms)

        return self._last_value


@function('HELD')
class HeldFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = {'asap'}

    STATE_OFF = 0
    STATE_WAITING = 1
    STATE_ON = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._start_time_ms: Optional[int] = None
        self._state = self.STATE_OFF

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, fixed_value, duration = await self.eval_args(context)

        if value == fixed_value:
            if self._state == self.STATE_OFF:
                self._start_time_ms = context.now_ms
                self._state = self.STATE_WAITING
                self.pause_asap_eval(self._start_time_ms + duration)
            elif self._state == self.STATE_WAITING:
                delta = context.now_ms - self._start_time_ms
                if delta >= duration:
                    self._state = self.STATE_ON
                self.pause_asap_eval()
        else:
            self._state = self.STATE_OFF
            self.pause_asap_eval()

        return self._state == self.STATE_ON


@function('DERIV')
class DerivFunction(Function):
    MIN_ARGS = MAX_ARGS = 2
    DEPS = {'asap'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_time_ms: int = 0

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, sampling_interval = await self.eval_args(context)
        result = 0

        if self._last_value is not None:
            delta = context.now_ms - self._last_time_ms
            if delta < sampling_interval:
                self.pause_asap_eval(self._last_time_ms + sampling_interval)
                raise EvalSkipped()

            if delta > TIME_JUMP_THRESHOLD:
                self._last_value = value
                self._last_time_ms = context.now_ms
                raise EvalSkipped()

            result = (value - self._last_value) / delta * 1000

        self._last_value = value
        self._last_time_ms = context.now_ms

        return result


@function('INTEG')
class IntegFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = {'asap'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: Optional[float] = None
        self._last_time_ms: int = 0

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, accumulator, sampling_interval = await self.eval_args(context)
        result = accumulator

        if self._last_value is not None:
            delta = context.now_ms - self._last_time_ms
            if delta < sampling_interval:
                self.pause_asap_eval(self._last_time_ms + sampling_interval)
                raise EvalSkipped()

            if delta > TIME_JUMP_THRESHOLD:
                self._last_value = value
                self._last_time_ms = context.now_ms
                raise EvalSkipped()

            result += (value + self._last_value) * delta / 2000

        self._last_value = value
        self._last_time_ms = context.now_ms

        return result


@function('FMAVG')
class FMAvgFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = {'asap'}
    QUEUE_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: list[float] = []
        self._last_time_ms: int = 0

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, width, sampling_interval = await self.eval_args(context)
        width = min(width, self.QUEUE_SIZE)

        if self._last_time_ms > 0:
            delta = context.now_ms - self._last_time_ms
            if delta < sampling_interval:
                self.pause_asap_eval(self._last_time_ms + sampling_interval)
                raise EvalSkipped()

            if delta > TIME_JUMP_THRESHOLD:
                self._last_time_ms = context.now_ms
                raise EvalSkipped()

        # Make place for the new element
        while len(self._queue) >= width:
            self._queue.pop(0)

        self._queue.append(value)
        self._last_time_ms = context.now_ms

        queue = self._queue[-int(width):]

        return sum(queue) / len(queue)


@function('FMEDIAN')
class FMedianFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = {'asap'}
    QUEUE_SIZE = 1024

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._queue: list[float] = []
        self._last_time_ms: int = 0

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, width, sampling_interval = await self.eval_args(context)
        width = min(width, self.QUEUE_SIZE)

        if self._last_time_ms > 0:
            delta = context.now_ms - self._last_time_ms
            if delta < sampling_interval:
                self.pause_asap_eval(self._last_time_ms + sampling_interval)
                raise EvalSkipped()

            if delta > TIME_JUMP_THRESHOLD:
                self._last_time_ms = context.now_ms
                raise EvalSkipped()

        # Make place for the new element
        while len(self._queue) >= width:
            self._queue.pop(0)

        self._queue.append(value)
        self._last_time_ms = context.now_ms

        queue = self._queue[-int(width):]
        queue.sort()

        return queue[len(queue) // 2]



from __future__ import annotations

import abc
import asyncio
import functools
import logging

from typing import Any, Callable, Optional

import bleak

from dbus_next import BusType, Variant
from dbus_next.aio import MessageBus
from dbus_next.errors import DBusError
from dbus_next.service import ServiceInterface, method

from qtoggleserver.core import ports as core_ports

from . import polled


BLUEZ_PROPERTIES_IFACE = 'org.freedesktop.DBus.Properties'
BLUEZ_BUS_NAME = 'org.bluez'
BLUEZ_DEVICE_IFACE = 'org.bluez.Device1'
BLUEZ_AGENT_PATH = '/io/qtoggle/agent'
BLUEZ_AGENT_IFACE = 'org.bluez.Agent1'
BLUEZ_AGENT_MANAGER_IFACE = 'org.bluez.AgentManager1'
BLUEZ_AGENT_MANAGER_PATH = '/org/bluez'
BLUEZ_AGENT_CAPABILITY = 'KeyboardDisplay'
BLUEZ_ADAPTER_PATH = '/org/bluez/hci0'
BLUEZ_ADAPTER_IFACE = 'org.bluez.Adapter1'

logger = logging.getLogger(__name__)

agent: Optional[Agent] = None


class BLEException(Exception):
    pass


class CommandTimeout(BLEException):
    pass


class NotificationTimeout(BLEException):
    pass


class Agent(ServiceInterface):
    def __init__(self, bus: MessageBus) -> None:
        super().__init__(BLUEZ_AGENT_IFACE)

        self.bus: MessageBus = bus
        self.secrets_by_address: dict[str, Optional[str]] = {}

    async def set_trusted(self, path: str) -> None:
        logger.debug('setting device at BT path %s as trusted', path)

        introspection = await self.bus.introspect(BLUEZ_BUS_NAME, path)
        obj = self.bus.get_proxy_object(BLUEZ_BUS_NAME, path, introspection)
        props = obj.get_interface(BLUEZ_PROPERTIES_IFACE)
        await props.call_set(BLUEZ_DEVICE_IFACE, 'Trusted', Variant('b', True))

    def add_device(self, address: str, secret: Optional[str]) -> None:
        self.secrets_by_address[address.lower()] = secret

    def rem_device(self, address: str) -> None:
        self.secrets_by_address.pop(address, None)

    @method()
    async def Release(self):  # noqa
        logger.debug('BT agent Release() called')

    @method()
    async def Cancel(self):  # noqa
        logger.debug('BT agent Cancel() called')

    @method()
    async def AuthorizeService(self, device: 'o', uuid: 's'):  # noqa
        logger.debug('BT agent AuthorizeService(%s, %s) called', device, uuid)
        address = self.address_from_device_path(device)
        if address not in self.secrets_by_address:
            logger.warning('got BT service authorization request from unknown device %s', address)
            raise DBusError('org.bluez.Error.Rejected', 'unknown device')

    @method()
    async def RequestConfirmation(self, device: 'o', passkey: 'u'):  # noqa
        logger.debug('BT agent RequestConfirmation(%s, %s) called', device, passkey)

        _none = {}
        address = self.address_from_device_path(device)
        secret = self.secrets_by_address.get(address, _none)
        if secret is _none:
            logger.warning('got BT request confirmation from unknown device %s', address)
            raise DBusError('org.bluez.Error.Rejected', 'unknown device')

        if not secret:
            logger.warning('got BT passkey request from device %s with unconfigured secret', address)
            raise DBusError('org.bluez.Error.Rejected', 'unconfigured secret')

        # `passkey` is a 0-padded 6-digit string
        while len(secret) < 6:
            secret = '0' + secret
        if secret != passkey:
            logger.warning('got BT request confirmation from device %s with invalid passkey %s', address, passkey)
            raise DBusError('org.bluez.Error.Rejected', 'invalid passkey')

    @method()
    async def RequestAuthorization(self, device: 'o'):  # noqa
        logger.debug('BT agent RequestAuthorization(%s) called', device)
        address = self.address_from_device_path(device)
        if address not in self.secrets_by_address:
            logger.warning('got BT authorization request from unknown device %s', address)
            raise DBusError('org.bluez.Error.Rejected', 'unknown device')

    @method()
    async def RequestPasskey(self, device: 'o') -> 'u':  # noqa
        logger.debug('BT agent RequestPasskey(%s) called', device)

        _none = {}
        address = self.address_from_device_path(device)
        secret = self.secrets_by_address.get(address, _none)
        if secret is _none:
            logger.warning('got BT passkey request from unknown device %s', address)
            raise DBusError('org.bluez.Error.Rejected', 'unknown device')

        if not secret:
            logger.warning('got BT passkey request from device %s with unconfigured secret', address)
            raise DBusError('org.bluez.Error.Rejected', 'unconfigured secret')

        await self.set_trusted(device)
        return int(secret)

    @method()
    async def RequestPinCode(self, device: 'o') -> 's':  # noqa
        logger.debug('BT agent RequestPinCode(%s) called', device)

        _none = {}
        address = self.address_from_device_path(device)
        secret = self.secrets_by_address.get(address, _none)
        if secret is _none:
            logger.warning('got BT pin code request from unknown device %s', address)
            raise DBusError('org.bluez.Error.Rejected', 'unknown device')

        if not secret:
            logger.warning('got BT passkey request from device %s with unconfigured secret', address)
            raise DBusError('org.bluez.Error.Rejected', 'unconfigured secret')

        await self.set_trusted(device)
        return secret

    @method()
    async def DisplayPasskey(self, device: 'o', passkey: 'u', entered: 'q'):  # noqa
        logger.debug('BT agent DisplayPasskey(%s, %s, %s) called', device, passkey, entered)

    @method()
    async def DisplayPinCode(self, device: 'o', pincode: 's'):  # noqa
        logger.debug('BT agent DisplayPinCode(%s, %s) called', device, pincode)

    @staticmethod
    def address_from_device_path(device_path: str) -> str:
        return ':'.join(device_path.split('_')[-6:]).lower()


async def get_agent() -> Agent:
    global agent

    if agent is None:
        logger.debug('initializing BT agent')
        bus = await MessageBus(bus_type=BusType.SYSTEM).connect()
        agent = Agent(bus)

        agent = Agent(bus)
        bus.export(BLUEZ_AGENT_PATH, agent)

        # Register agent
        introspection = await bus.introspect(BLUEZ_BUS_NAME, BLUEZ_AGENT_MANAGER_PATH)
        obj = bus.get_proxy_object(BLUEZ_BUS_NAME, BLUEZ_AGENT_MANAGER_PATH, introspection)
        agent_manager_obj = obj.get_interface(BLUEZ_AGENT_MANAGER_IFACE)

        await agent_manager_obj.call_register_agent(BLUEZ_AGENT_PATH, BLUEZ_AGENT_CAPABILITY)
        await agent_manager_obj.call_request_default_agent(BLUEZ_AGENT_PATH)

        # Make adapter discoverable/pairable
        introspection = await bus.introspect(BLUEZ_BUS_NAME, BLUEZ_ADAPTER_PATH)
        obj = bus.get_proxy_object(BLUEZ_BUS_NAME, BLUEZ_ADAPTER_PATH, introspection)
        props = obj.get_interface(BLUEZ_PROPERTIES_IFACE)

        await props.call_set(BLUEZ_ADAPTER_IFACE, 'DiscoverableTimeout', Variant('u', 0))
        await props.call_set(BLUEZ_ADAPTER_IFACE, 'Discoverable', Variant('b', True))
        await props.call_set(BLUEZ_ADAPTER_IFACE, 'PairableTimeout', Variant('u', 0))
        await props.call_set(BLUEZ_ADAPTER_IFACE, 'Pairable', Variant('b', True))

    return agent


class BLEPeripheral(polled.PolledPeripheral, metaclass=abc.ABCMeta):
    DEFAULT_CMD_TIMEOUT = 30
    DEFAULT_RETRY_COUNT = 3
    DEFAULT_RETRY_DELAY = 5
    TIMEOUT_ATOM = 0.1

    logger = logger

    def __init__(
        self,
        *,
        address: str,
        secret: Optional[str] = None,
        cmd_timeout: int = DEFAULT_CMD_TIMEOUT,
        retry_count: int = DEFAULT_RETRY_COUNT,
        retry_delay: int = DEFAULT_RETRY_DELAY,
        **kwargs
    ) -> None:
        self._address: str = address
        self._secret: Optional[str] = secret
        self._cmd_timeout: int = cmd_timeout
        self._retry_count: int = retry_count
        self._retry_delay: int = retry_delay
        self._busy: bool = False
        self._notification_data: Optional[bytes] = None

        super().__init__(**kwargs)

    def __str__(self) -> str:
        return self.get_name()

    async def handle_init(self) -> None:
        agent = await get_agent()
        agent.add_device(self._address, self._secret)

    async def handle_cleanup(self) -> None:
        agent = await get_agent()
        agent.rem_device(self._address)

    async def read(self, handle: int) -> bytes:
        return await self._run_cmd(self._read, handle=handle)

    async def write(self, handle: int, data: bytes) -> None:
        return await self._run_cmd(self._write, handle=handle, data=data)

    async def wait_notify(self, handle: int) -> bytes:
        return await self._run_cmd(self._wait_notify, handle=handle)

    async def write_wait_notify(self, handle: int, notify_handle: int, data: bytes) -> bytes:
        return await self._run_cmd(self._write_wait_notify, handle=handle, data=data, notify_handle=notify_handle)

    async def _run_cmd(self, cmd_func: Callable, **kwargs) -> Optional[bytes]:
        # Wait for the peripheral to be ready (not busy)
        timeout = self._cmd_timeout
        while self._busy and timeout > 0:
            await asyncio.sleep(self.TIMEOUT_ATOM)
            timeout -= self.TIMEOUT_ATOM

        if timeout <= 0:
            raise CommandTimeout()

        self._busy = True
        retry = 1
        retry_count = self._retry_count
        while True:
            try:
                response = await asyncio.wait_for(cmd_func(timeout=timeout, **kwargs), timeout)
            except Exception:
                self.error('command execution failed', exc_info=True)

                if retry <= retry_count:
                    await asyncio.sleep(self._retry_delay)
                    self.debug('retry %s/%s', retry, retry_count)
                    retry += 1
                    continue

                self.set_online(False)
                self._busy = False
                raise
            else:
                self.set_online(True)
                self._busy = False
                return response

    async def _read(self, handle: int, timeout: int) -> bytes:
        self.debug('connecting')
        async with bleak.BleakClient(self._address, timeout=timeout) as client:
            self.debug('reading from %04X', handle)
            response = bytes(await client.read_gatt_char(handle))
            self.debug('got response: %s', self.pretty_data(response))
        self.debug('disconnected')

        return response

    async def _write(self, handle: int, data: bytes, timeout: int) -> None:
        self.debug('connecting')
        async with bleak.BleakClient(self._address, timeout=timeout) as client:
            self.debug('writing at %04X: %s', handle, self.pretty_data(data))
            await client.write_gatt_char(handle, data)
        self.debug('disconnected')

    async def _wait_notify(self, handle: int, timeout: int) -> bytes:
        self.debug('connecting')
        async with bleak.BleakClient(self._address, timeout=timeout) as client:
            self.debug('waiting for notification on %04X', handle)
            self._notification_data = None
            await client.start_notify(handle, self._notify_callback)
            try:
                for _ in range(int(timeout / self.TIMEOUT_ATOM)):
                    await asyncio.sleep(self.TIMEOUT_ATOM)
                    if self._notification_data:
                        break
                else:
                    raise NotificationTimeout()
            finally:
                await client.stop_notify(handle)
        self.debug('disconnected')

        return self._notification_data

    async def _write_wait_notify(self, handle: int, notify_handle: int, data: bytes, timeout: int) -> bytes:
        self.debug('connecting')
        async with bleak.BleakClient(self._address, timeout=timeout) as client:
            self.debug('writing at %04X: %s', handle, self.pretty_data(data))
            await client.write_gatt_char(handle, data)
            self.debug('waiting for notification on %04X', notify_handle)
            self._notification_data = None
            await client.start_notify(notify_handle, self._notify_callback)
            try:
                for _ in range(int(timeout / self.TIMEOUT_ATOM)):
                    await asyncio.sleep(self.TIMEOUT_ATOM)
                    if self._notification_data:
                        break
                else:
                    raise NotificationTimeout()
            finally:
                await client.stop_notify(notify_handle)
        self.debug('disconnected')

        return self._notification_data

    def _notify_callback(self, handle: int, data: bytearray) -> None:
        self._notification_data = bytes(data)
        self.debug('got notification on %04X: %s', handle, self.pretty_data(self._notification_data))

    @staticmethod
    def pretty_data(data: bytes) -> str:
        return ' '.join(map(lambda c: f'{c:02X}', data))


class BLEPort(polled.PolledPort, metaclass=abc.ABCMeta):
    READ_INTERVAL_MAX = 1440
    READ_INTERVAL_STEP = 1
    READ_INTERVAL_MULTIPLIER = 60


def port_exceptions(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Any:
        # Transform BLE exceptions into port exceptions, where applicable
        try:
            return func(*args, **kwargs)
        except (NotificationTimeout, CommandTimeout) as e:
            raise core_ports.PortTimeout() from e
        except BLEException as e:
            raise core_ports.PortError(str(e)) from e

    return wrapper



from __future__ import annotations

import asyncio
import logging
import time

from typing import Iterable, Optional

from qtoggleserver import persist, system
from qtoggleserver.conf import settings
from qtoggleserver.core import events as core_events
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import GenericJSONDict, PortValue
from qtoggleserver.utils import json as json_utils


PERSIST_COLLECTION = 'value_history'

_CACHE_TIMESTAMP_MIN_AGE = 3600 * 1000  # don't cache samples newer than this number of milliseconds ago

logger = logging.getLogger(__name__)

_history_event_handler: Optional[HistoryEventHandler] = None
_sampling_task: Optional[asyncio.Task] = None
_janitor_task: Optional[asyncio.Task] = None

# Samples cached by port_id and timestamp
_samples_cache: dict[str, dict[int, PortValue]] = {}

# Used to schedule sample removal with remove_samples(..., background=True)
_pending_remove_samples: list[tuple[core_ports.BasePort, Optional[int], Optional[int]]] = []


class HistoryEventHandler(core_events.Handler):
    FIRE_AND_FORGET = True

    async def handle_event(self, event: core_events.Event) -> None:
        if not isinstance(event, core_events.ValueChange):
            return  # we're only interested in port value changes

        if not system.date.has_real_date_time():
            return  # don't record history unless we've got real date/time

        port = event.get_port()
        history_interval = await port.get_history_interval()
        if history_interval != -1:
            return  # only consider ports with history interval set to special -1 (on value change)

        now_ms = int(time.time() * 1000)

        await save_sample(port, now_ms)
        port.set_history_last_timestamp(now_ms)


async def sampling_task() -> None:
    while True:
        try:
            await asyncio.sleep(1)

            if not system.date.has_real_date_time():
                continue  # don't record history unless we've got real date/time

            now_ms = int(time.time() * 1000)
            for port in core_ports.get_all():
                if not port.is_enabled():
                    continue

                history_last_timestamp = port.get_history_last_timestamp()
                history_interval = await port.get_history_interval()
                if history_interval <= 0:  # disabled or on value change
                    continue

                if now_ms - history_last_timestamp < history_interval * 1000:
                    continue

                await save_sample(port, now_ms)
                port.set_history_last_timestamp(now_ms)
                port.save_asap()  # history_last_timestamp must be persisted
        except asyncio.CancelledError:
            logger.debug('sampling task cancelled')
            break
        except Exception as e:
            logger.error('sampling task error: %s', e, exc_info=True)


async def janitor_task() -> None:
    global _pending_remove_samples

    while True:
        try:
            await asyncio.sleep(settings.core.history_janitor_interval)

            if not system.date.has_real_date_time():
                continue

            now = int(time.time())
            for port in core_ports.get_all():
                history_retention = await port.get_history_retention()
                if history_retention <= 0:
                    continue

                to_timestamp = (now - history_retention) * 1000
                logger.debug('removing old samples of %s from history', port)
                await remove_samples([port], from_timestamp=0, to_timestamp=to_timestamp)

            # Remove samples that were background-scheduled for removal
            # Group together requests w/o timestamp bounds

            ports = []
            rem_pending_remove_samples = []
            for port, from_timestamp, to_timestamp in _pending_remove_samples:
                if from_timestamp is to_timestamp is None:
                    ports.append(port)
                else:
                    rem_pending_remove_samples.append((port, from_timestamp, to_timestamp))

            if ports:
                port_ids = [p.get_id() for p in ports]
                logger.debug('removing samples of %s from history (background)', ', '.join(port_ids))
                await remove_samples(ports)

            _pending_remove_samples = rem_pending_remove_samples
            while _pending_remove_samples:
                port, from_timestamp, to_timestamp = _pending_remove_samples.pop(0)
                logger.debug('removing samples of %s from history (background)', port)
                await remove_samples([port], from_timestamp=from_timestamp, to_timestamp=to_timestamp)
        except asyncio.CancelledError:
            logger.debug('janitor task cancelled')
            break
        except Exception as e:
            logger.error('janitor task error: %s', e, exc_info=True)


def is_enabled() -> bool:
    return persist.is_history_supported() and settings.core.history_support


async def get_samples_slice(
    port: core_ports.BasePort,
    from_timestamp: Optional[int] = None,
    to_timestamp: Optional[int] = None,
    limit: Optional[int] = None,
    sort_desc: bool = False
) -> Iterable[GenericJSONDict]:
    filt = {
        'pid': port.get_id(),
    }

    if from_timestamp is not None:
        filt.setdefault('ts', {})['ge'] = from_timestamp

    if to_timestamp is not None:
        filt.setdefault('ts', {})['lt'] = to_timestamp

    sort = 'ts'
    if sort_desc:
        sort = f'-{sort}'

    results = await persist.query(PERSIST_COLLECTION, filt=filt, sort=sort, limit=limit)

    return ({'value': r['val'], 'timestamp': r['ts']} for r in results)


async def get_samples_by_timestamp(
    port: core_ports.BasePort,
    timestamps: list[int]
) -> Iterable[GenericJSONDict]:
    port_filter = {
        'pid': port.get_id(),
    }

    now_ms = int(time.time() * 1000)
    samples_cache = _samples_cache.setdefault(port.get_id(), {})
    INEXISTENT = {}

    query_tasks = []
    for timestamp in timestamps:
        # Look it up in cache
        sample = samples_cache.get(timestamp, INEXISTENT)
        if sample is INEXISTENT:
            filt = dict(port_filter, ts={'le': timestamp})
            task = persist.query(PERSIST_COLLECTION, filt=filt, sort='-ts', limit=1)
        else:
            task = asyncio.Future()
            task.set_result([sample])

        query_tasks.append(task)

    task_results = await asyncio.gather(*query_tasks)

    samples = []
    for i, task_result in enumerate(task_results):
        timestamp = timestamps[i]

        query_results = list(task_result)
        if query_results:
            sample = query_results[0]
            samples.append(sample)
        else:
            samples.append(None)

        # Add sample to cache if it's old enough
        if now_ms - timestamp > _CACHE_TIMESTAMP_MIN_AGE:
            samples_cache[timestamp] = samples[-1]

    return ({'value': r['val'], 'timestamp': r['ts']} if r is not None else None for r in samples)


async def save_sample(port: core_ports.BasePort, timestamp: int) -> None:
    value = port.get_last_read_value()

    if value is None:
        logger.debug('skipping null sample of %s (timestamp = %s)', port, timestamp)
        return

    logger.debug('saving sample of %s (value = %s, timestamp = %s)', port, json_utils.dumps(value), timestamp)

    record = {
        'pid': port.get_id(),
        'val': value,
        'ts': timestamp
    }

    await persist.insert(PERSIST_COLLECTION, record)


async def remove_samples(
    ports: list[core_ports.BasePort],
    from_timestamp: Optional[int] = None,
    to_timestamp: Optional[int] = None,
    background: bool = False
) -> Optional[int]:
    filt = {
        'pid': {'in': [p.get_id() for p in ports]}
    }

    if from_timestamp is not None:
        filt.setdefault('ts', {})['ge'] = from_timestamp

    if to_timestamp is not None:
        filt.setdefault('ts', {})['lt'] = to_timestamp

    # Invalidate samples cache for the ports
    for port in ports:
        _samples_cache.pop(port.get_id(), None)

    if background:
        for port in ports:
            _pending_remove_samples.append((port, from_timestamp, to_timestamp))
    else:
        return await persist.remove(PERSIST_COLLECTION, filt)


async def reset() -> None:
    logger.debug('clearing persisted data')
    await persist.remove(PERSIST_COLLECTION)


async def init() -> None:
    global _history_event_handler
    global _sampling_task
    global _janitor_task

    _history_event_handler = HistoryEventHandler()
    core_events.register_handler(_history_event_handler)

    _sampling_task = asyncio.create_task(sampling_task())
    _janitor_task = asyncio.create_task(janitor_task())

    await persist.ensure_index(PERSIST_COLLECTION, 'ts')


async def cleanup() -> None:
    if _sampling_task:
        _sampling_task.cancel()
        await _sampling_task

    if _janitor_task:
        _janitor_task.cancel()
        await _janitor_task



import logging
import os
import subprocess

import psutil

from qtoggleserver.conf import settings

from . import battery, conf, date, dns, fwupdate, net, storage, temperature


logger = logging.getLogger(__name__)


def is_setup_mode() -> bool:
    if not settings.system.setup_mode_cmd:
        return False

    try:
        subprocess.check_call(settings.system.setup_mode_cmd, stderr=subprocess.STDOUT, shell=True)
        return True
    except subprocess.CalledProcessError:
        return False


def reboot() -> None:
    logger.debug('rebooting')
    if not settings.debug:
        os.system('reboot')


def uptime() -> int:
    with open('/proc/uptime', 'r') as f:
        line = f.readlines()[0]
        return int(float(line.split()[0]))


def get_cpu_usage() -> int:
    return int(psutil.cpu_percent())


def get_mem_usage() -> int:
    vm = psutil.virtual_memory()
    return int(100 * (1 - vm.available / vm.total))


async def init() -> None:
    logger.debug('initializing DNS')
    await dns.init()


async def cleanup() -> None:
    logger.debug('cleaning up DNS')
    await dns.cleanup()



import re

from typing import Optional

from qtoggleserver.core.typing import NullablePortValue as CoreNullablePortValue

from .base import EvalContext, EvalResult, Expression
from .exceptions import EmptyExpression, UnexpectedCharacter, ValueUnavailable


class LiteralValue(Expression):
    def __init__(self, value: CoreNullablePortValue, sexpression: str, role: int) -> None:
        super().__init__(role)

        self.value: CoreNullablePortValue = value
        self.sexpression: str = sexpression

    def __str__(self) -> str:
        return self.sexpression

    async def _eval(self, context: EvalContext) -> EvalResult:
        if self.value is None:
            raise ValueUnavailable

        return float(self.value)

    @staticmethod
    def parse(self_port_id: Optional[str], sexpression: str, role: int, pos: int) -> Expression:
        while sexpression and sexpression[0].isspace():
            sexpression = sexpression[1:]
            pos += 1

        while sexpression and sexpression[-1].isspace():
            sexpression = sexpression[:-1]

        if not sexpression:
            raise EmptyExpression()

        if sexpression == 'true':
            value = 1
        elif sexpression == 'false':
            value = 0
        elif sexpression == 'unavailable':
            value = None
        else:
            try:
                value = int(sexpression)
            except ValueError:
                try:
                    value = float(sexpression)
                except ValueError:
                    m = re.match(r'-?\d+(\.?\d+)?', sexpression)
                    if m:
                        raise UnexpectedCharacter(sexpression[m.end()], pos + m.end()) from None
                    else:
                        raise UnexpectedCharacter(sexpression[0], pos) from None

        return LiteralValue(value, sexpression, role)



import hashlib
import logging
import re
import time

from typing import Callable, Optional

import jwt

from qtoggleserver import system
from qtoggleserver.conf import settings
from qtoggleserver.core.device import attrs as core_device_attrs


JWT_ISS = 'qToggle'
JWT_ALG = 'HS256'

ORIGIN_DEVICE = 'device'
ORIGIN_CONSUMER = 'consumer'

EMPTY_PASSWORD_HASH = hashlib.sha256(b'').hexdigest()

_AUTH_TOKEN_RE = re.compile(r'^Bearer\s+([a-z0-9_.-]+)$', re.IGNORECASE)

logger = logging.getLogger(__name__)


class AuthError(Exception):
    pass


def make_auth_header(origin: str, username: Optional[str], password_hash: str) -> str:
    claims = {
        'iss': JWT_ISS,
        'ori': origin
    }

    if username:
        claims['usr'] = username

    if system.date.has_real_date_time():
        claims['iat'] = int(time.time())

    token = jwt.encode(claims, key=password_hash or '', algorithm=JWT_ALG)

    return f'Bearer {token}'


def parse_auth_header(auth: str, origin: str, password_hash_func: Callable, require_usr: bool = True) -> str:
    m = _AUTH_TOKEN_RE.match(auth)
    if not m:
        raise AuthError('Invalid authorization header')

    # Decode but don't validate token yet
    token = m.group(1)

    try:
        payload = jwt.decode(
            token, algorithms=[JWT_ALG], options={'verify_signature': False}, leeway=settings.core.max_client_time_skew
        )
    except jwt.exceptions.InvalidTokenError as e:
        raise AuthError(f'Invalid JWT: {e}') from e

    # Validate claims
    if payload.get('iss') != JWT_ISS:
        raise AuthError('Missing or invalid iss claim in JWT')

    if payload.get('ori') != origin:
        raise AuthError('Missing or invalid ori claim in JWT')

    iat = payload.get('iat')
    if (iat is not None) and system.date.has_real_date_time():
        delta = time.time() - iat
        if abs(delta) > settings.core.max_client_time_skew:
            raise AuthError('JWT too old or too much in the future')

    usr = payload.get('usr')
    if require_usr:
        if not usr or not isinstance(usr, str):
            raise AuthError('Missing or invalid usr claim in JWT')

    # Validate username & signature
    password_hash = password_hash_func(usr)
    if not password_hash:
        raise AuthError(f'Unknown usr in JWT: {usr}')

    # Decode again to verify signature
    try:
        jwt.decode(
            token, key=password_hash, algorithms=[JWT_ALG], verify=True, leeway=settings.core.max_client_time_skew
        )
    except jwt.exceptions.InvalidSignatureError as e:
        raise AuthError('Invalid JWT signature') from e
    except jwt.exceptions.InvalidTokenError as e:
        raise AuthError(f'Invalid JWT: {e}') from e

    return usr


def consumer_password_hash_func(usr: str) -> Optional[str]:
    if usr == 'admin':
        return core_device_attrs.admin_password_hash
    elif usr == 'normal':
        return core_device_attrs.normal_password_hash
    elif usr == 'viewonly':
        return core_device_attrs.viewonly_password_hash
    else:
        return None



from .base import EvalContext, EvalResult
from .exceptions import ExpressionArithmeticError
from .functions import Function, function


@function('ADD')
class AddFunction(Function):
    MIN_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        return sum(await self.eval_args(context))


@function('SUB')
class SubFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)
        return eval_args[0] - eval_args[1]


@function('MUL')
class MulFunction(Function):
    MIN_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        r = 1
        for e in await self.eval_args(context):
            r *= e

        return r


@function('DIV')
class DivFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        if eval_args[1]:
            return eval_args[0] / eval_args[1]
        else:
            raise ExpressionArithmeticError


@function('MOD')
class ModFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        if eval_args[1]:
            return eval_args[0] % eval_args[1]
        else:
            raise ExpressionArithmeticError


@function('POW')
class PowFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        eval_args = await self.eval_args(context)

        return eval_args[0] ** eval_args[1]



from __future__ import annotations

import abc

from typing import Optional, Union

from .exceptions import EvalSkipped, ExpressionEvalError


class Expression(metaclass=abc.ABCMeta):
    def __init__(self, role: int) -> None:
        self.role: int = role
        self._asap_eval_paused_until_ms: int = 0
        self._cached_deps: Optional[set[str]] = None

    def pause_asap_eval(self, pause_until_ms: int = 0) -> None:
        self._asap_eval_paused_until_ms = pause_until_ms or int(1e13)

    def is_asap_eval_paused(self, now_ms: int) -> bool:
        return now_ms < self._asap_eval_paused_until_ms

    async def eval(self, context: EvalContext) -> EvalResult:
        self._asap_eval_paused_until_ms = 0
        try:
            return await self._eval(context)
        except EvalSkipped:
            raise
        except ExpressionEvalError:
            # Pause expression evaluation for 1 second, as it's very unlikely that a problematic expression become
            # fixed within a second. This is a small speed optimization for expressions that depend on millisecond.
            self.pause_asap_eval(context.now_ms + 1000)
            raise

    @abc.abstractmethod
    async def _eval(self, context: EvalContext) -> EvalResult:
        raise NotImplementedError()

    def get_deps(self) -> set[str]:
        if self._cached_deps is None:
            self._cached_deps = self._get_deps()
        return self._cached_deps

    def _get_deps(self) -> set[str]:
        # Special deps:
        #  * 'second' - used to indicate dependency on system time (seconds)
        #  * 'asap' - used to indicate that evaluation should be done as soon as possible
        return set()

    @staticmethod
    @abc.abstractmethod
    def parse(self_port_id: Optional[str], sexpression: str, role: int, pos: int) -> Expression:
        raise NotImplementedError()


class EvalContext:
    def __init__(
        self,
        port_values: dict[str, NullablePortValue],
        now_ms: int
    ) -> None:
        self.port_values: dict[str, NullablePortValue] = port_values
        self.now_ms: int = now_ms

    @property
    def timestamp(self) -> int:
        return int(self.now_ms / 1000)


# This needs to be imported here to avoid circular import issues
from qtoggleserver.core import ports as core_ports  # noqa: E402
from qtoggleserver.core.typing import NullablePortValue  # noqa: E402


EvalResult = Union[bool, int, float, core_ports.BasePort]



import abc
import asyncio
import inspect
import queue
import sys
import threading

from typing import Any, Awaitable, Callable, Optional, Union


class ParallelCaller:
    def __init__(self, parallel: int = 1, max_queued: int = 0) -> None:
        self._queue: asyncio.Queue = asyncio.Queue(max_queued)
        self._loop_tasks: list[Optional[asyncio.Task]] = []

        # Start loop tasks
        for i in range(parallel):
            self._loop_tasks.append(asyncio.create_task(self._loop(i)))

    async def call(self, func: Callable, *args, is_async: Optional[bool] = None, **kwargs) -> Any:
        # Push call details to queue
        when_ready = asyncio.Condition()
        result = {'ret': None, 'exc_info': (None, None, None)}

        # Allow QueueFull exceptions to be propagated to caller
        self._queue.put_nowait((func, is_async, args, kwargs, when_ready, result))

        # Wait for call to be ready
        async with when_ready:
            await when_ready.wait()

        # If an exception was raised, re-raise it
        typ, val, tb = result['exc_info']
        if typ:
            raise val

        return result['ret']

    async def _loop(self, index: int) -> None:
        try:
            while True:
                func, is_async, args, kwargs, when_ready, result = await self._queue.get()

                if is_async is None:
                    is_async = inspect.iscoroutinefunction(func)

                try:
                    if is_async:
                        result['ret'] = await func(*args, **kwargs)
                    else:
                        result['ret'] = func(*args, **kwargs)
                except Exception:
                    result['exc_info'] = sys.exc_info()

                async with when_ready:
                    when_ready.notify_all()
        except asyncio.CancelledError:
            pass

    async def stop(self) -> None:
        current_task = asyncio.current_task()

        for task in self._loop_tasks:
            if task is not current_task:
                task.cancel()

        await asyncio.gather(*[t for t in self._loop_tasks if t is not current_task])


class Timer:
    def __init__(self, timeout: int, callback: Union[Awaitable, Callable], *args, **kwargs) -> None:
        self._timeout: int = timeout
        self._callback: Union[Awaitable, Callable] = callback
        self._args: tuple = args
        self._kwargs: dict = kwargs
        self._task: Optional[asyncio.Task] = None
        self._task = asyncio.ensure_future(self.run())

    def cancel(self) -> None:
        if self._task is None:
            raise Exception('Task is not running')

        self._task.cancel()

    async def wait(self) -> None:
        if self._task is None:
            raise Exception('Task is not running')

        await self._task

    async def run(self) -> None:
        try:
            await asyncio.sleep(self._timeout)
        except asyncio.CancelledError:
            pass
        else:
            result = self._callback(*self._args, **self._kwargs)
            if inspect.isawaitable(result):
                await result
        finally:
            self._task = None


class RunnerBusy(Exception):
    pass


class ThreadedRunner(threading.Thread, metaclass=abc.ABCMeta):
    QUEUE_TIMEOUT = 1

    def __init__(self, queue_size: Optional[int] = None) -> None:
        self._running: bool = False
        self._loop: asyncio.AbstractEventLoop = asyncio.get_running_loop()
        self._queue: queue.Queue = queue.Queue(queue_size or 0)
        self._queue_size: int = queue_size
        self._stopped_future: asyncio.Future = self._loop.create_future()

        super().__init__()

    def run(self) -> None:
        while self._running:
            try:
                func, callback = self._queue.get(timeout=self.QUEUE_TIMEOUT)
            except queue.Empty:
                continue

            try:
                result = func()
            except Exception as e:
                self._loop.call_soon_threadsafe(callback, None, e)
            else:
                self._loop.call_soon_threadsafe(callback, result, None)

        self._loop.call_soon_threadsafe(self._stopped_future.set_result, None)

    def schedule_func(self, func: Callable, callback: Callable) -> None:
        try:
            self._queue.put_nowait((func, callback))
        except queue.Full:
            raise RunnerBusy() from None

    def is_running(self) -> bool:
        return self._running

    def start(self) -> None:
        self._running = True
        super().start()

    async def stop(self) -> None:
        self._running = False

        await self._stopped_future


async def await_later(delay: float, coroutine: Callable, *args, loop: asyncio.AbstractEventLoop = None) -> None:
    await asyncio.sleep(delay, loop=loop)
    await coroutine(*args)



import asyncio

from typing import Optional

from tornado import process as tornado_process


async def call_subprocess(args: list[str], stdin_data: Optional[bytes] = None) -> tuple[int, bytes, bytes]:
    p = tornado_process.Subprocess(
        args,
        stdin=tornado_process.Subprocess.STREAM,
        stdout=tornado_process.Subprocess.STREAM,
        stderr=tornado_process.Subprocess.STREAM
    )

    if stdin_data:
        await p.stdin.write(stdin_data)
        p.stdin.close()

    exit_future = p.wait_for_exit(raise_error=False)
    stdout_future = p.stdout.read_until_close()
    stderr_future = p.stderr.read_until_close()

    # TODO: these might need to be wrapped in asyncio.create_task()
    await asyncio.wait({exit_future, stdout_future, stderr_future})

    return exit_future.result(), stdout_future.result(), stderr_future.result()



import asyncio
import logging
import time

from typing import Optional, Union

from qtoggleserver.conf import settings
from qtoggleserver.core import expressions as core_expressions  # noqa: F401; Required to avoid circular imports
from qtoggleserver.core import ports as core_ports
from qtoggleserver.core.typing import NullablePortValue
from qtoggleserver.utils import json as json_utils
from qtoggleserver.utils import logging as logging_utils
from qtoggleserver.utils import timedset


# After how much time to retry reading a port whose read_value() method raised an error
_PORT_READ_ERROR_RETRY_INTERVAL = 10

logger = logging.getLogger(__name__)
memory_logs: Optional[logging_utils.FifoMemoryHandler] = None

loop: Optional[asyncio.AbstractEventLoop] = None

_update_loop_task: Optional[asyncio.Task] = None
_ready: bool = False
_updating_enabled: bool = True
_start_time: float = time.time()
_last_time: int = 0
_force_eval_expression_ports: set[core_ports.BasePort] = set()
_force_eval_all_expressions: bool = False
_ports_with_read_error = timedset.TimedSet(_PORT_READ_ERROR_RETRY_INTERVAL)
_update_lock: Optional[asyncio.Lock] = None


async def update() -> None:
    from . import ports, sessions

    global _last_time
    global _update_lock

    if not _updating_enabled:
        return

    if _update_lock is None:
        _update_lock = asyncio.Lock()

    async with _update_lock:
        changed_set: set[Union[core_ports.BasePort, str]] = {'asap'}
        value_pairs = {}

        now = time.time()
        now_int = int(now)
        second_changed = False
        if now_int != _last_time:
            _last_time = now_int
            second_changed = True
            changed_set.add('second')

        for port in ports.get_all():
            if not port.is_enabled():
                continue

            port.invalidate_attrs()
            old_value = port.get_last_read_value()

            if second_changed:
                try:
                    port.heart_beat_second()
                except Exception as e:
                    logger.error('port heart beat second exception: %s', e, exc_info=True)

            # Skip ports with read errors for a while
            if port in _ports_with_read_error:
                continue

            try:
                new_value = await port.read_transformed_value()
            except Exception as e:
                logger.error('failed to read value from %s: %s', port, e, exc_info=True)
                _ports_with_read_error.add(port)

                continue

            if new_value is None:
                continue  # read value not available

            if new_value != old_value:
                old_value_str = json_utils.dumps(old_value) if old_value is not None else '(unavailable)'
                new_value_str = json_utils.dumps(new_value)

                logger.debug('detected %s value change: %s -> %s', port, old_value_str, new_value_str)

                port.set_last_read_value(new_value)
                changed_set.add(port)
                value_pairs[port] = old_value, new_value

        await handle_value_changes(changed_set, value_pairs, now)

        sessions.update()


async def update_loop() -> None:
    while True:
        try:
            try:
                if _ready:
                    await update()
            except Exception as e:
                logger.error('update failed: %s', e, exc_info=True)
            await asyncio.sleep(settings.core.tick_interval / 1000.0)
        except asyncio.CancelledError:
            logger.debug('update loop cancelled')
            break


async def handle_value_changes(
    changed_set: set[Union[core_ports.BasePort, str]],
    value_pairs: dict[core_ports.BasePort, tuple[NullablePortValue, NullablePortValue]],
    now: float,
) -> None:
    global _force_eval_all_expressions

    # changed_set contains:
    #  * ports
    #  * time strings

    # deps contain:
    #  * `$`-prefixed port ids
    #  * time strings

    forced_ports = set(_force_eval_expression_ports)
    _force_eval_expression_ports.clear()

    full_eval = _force_eval_all_expressions
    _force_eval_all_expressions = False

    now_ms = int(now * 1000)

    # Transform `changed_set` into a set of strings so that we can compare it with deps
    changed_set_str = set()

    # Trigger value-change events; save persisted ports; build changed_set_str
    for port in changed_set:
        if isinstance(port, core_ports.BasePort):
            changed_set_str.add(f'${port.get_id()}')
        else:
            changed_set_str.add(port)
            continue

        if not await port.is_internal():
            value_pair = value_pairs.get(port)
            if not value_pair:
                continue
            await port.trigger_value_change(*value_pair)

        if await port.is_persisted():
            port.save_asap()

    # Reevaluate the expressions depending on changed ports
    for port in core_ports.get_all():
        if not port.is_enabled():
            continue

        expression = port.get_expression()
        if not expression:
            continue

        if full_eval or (port in forced_ports):
            port.push_eval()
            continue

        port_own_deps: set[str] = expression.get_deps()
        deps: set[str] = port_own_deps - {f'${port.get_id()}'}

        # Evaluate a port's expression only if one of its deps changed
        changed_deps = deps & changed_set_str
        if not changed_deps:
            continue

        if ('asap' in deps) and (len(changed_deps) == 1):
            # Skip asap evaling if explicitly paused
            if expression.is_asap_eval_paused(now_ms):
                continue

            # Don't flood port with evals
            if port.has_pending_eval():
                continue

        port.push_eval()


def force_eval_expressions(port: core_ports.BasePort = None) -> None:
    global _force_eval_all_expressions

    logger.debug('forcing expression evaluation for %s', port or 'all ports')

    if port:
        _force_eval_expression_ports.add(port)
    else:
        _force_eval_all_expressions = True


def enable_updating() -> None:
    global _updating_enabled

    if not _updating_enabled:
        logger.debug('enabling update mechanism')
        _updating_enabled = True


def disable_updating() -> None:
    global _updating_enabled

    if _updating_enabled:
        logger.debug('disabling update mechanism')
        _updating_enabled = False


def is_ready() -> bool:
    return _ready


def set_ready() -> None:
    global _ready

    logger.debug('ready')
    _ready = True


def uptime() -> float:
    return time.time() - _start_time


async def init() -> None:
    global _update_loop_task
    global loop

    loop = asyncio.get_event_loop()

    force_eval_expressions()
    _update_loop_task = loop.create_task(update_loop())


async def cleanup() -> None:
    if _update_loop_task:
        _update_loop_task.cancel()
        await _update_loop_task



from .base import EvalContext, EvalResult
from .functions import Function, function


@function('ABS')
class AbsFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def _eval(self, context: EvalContext) -> EvalResult:
        return abs((await self.eval_args(context))[0])


@function('SGN')
class SgnFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def _eval(self, context: EvalContext) -> EvalResult:
        e = int((await self.eval_args(context))[0])
        if e > 0:
            return 1
        elif e < 0:
            return -1
        else:
            return 0
